#include <linux/circ_buf.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/usb/input.h>
#include <linux/hid.h>
#include <linux/spinlock.h>

MODULE_LICENSE("GPL");

// MUST BE POWER OF 2
#define MAX_INBUF_LEN 1024

static char m[]  = "\0\0\0\0abcdefghijklmnopqrstuvwxyz1234567890\n\0\0\t -=[]\\\0;'`,./";
static char sm[] = "\0\0\0\0ABCDEFGHIJKLMNOPQRSTUVWXYZ!@#$%^&*()\n\0\0\t _+{}|\0:\"~<>?";

DEFINE_SPINLOCK(buf_lock);
static struct circ_buf cardscanner_inbuf = {0};

static int cardscanner_probe(struct usb_interface *iface, const struct usb_device_id *id);
static void cardscanner_disconnect(struct usb_interface *intf);

static struct usb_device_id cardscanner_id_table [] = {
	{ USB_DEVICE(0x801, 0x1) },
	{ }  
};

static struct usb_driver cardscanner_driver = {
	.name =		"cardscanner",
	.probe =	cardscanner_probe,
	.disconnect =	cardscanner_disconnect,
	.id_table =	cardscanner_id_table
};

static int cardscanner_open(struct inode *inode, struct file *filp)
{
	return 0;
}

static ssize_t cardscanner_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
    int head, tail;
    size_t i, tmp, newcount;
    unsigned long flags;
    char tmp_buf[MAX_INBUF_LEN];

    spin_lock_irqsave(&buf_lock, flags);
    head = cardscanner_inbuf.head;
    tail = cardscanner_inbuf.tail;

    tmp = CIRC_CNT(head, tail, MAX_INBUF_LEN);
    if (count > tmp)
        count = tmp;

    tmp = CIRC_CNT_TO_END(head, tail, MAX_INBUF_LEN);

    if (tmp > count)
        tmp = count;

    for (i = 0; i < tmp; ++i)
        tmp_buf[i] = cardscanner_inbuf.buf[tail + i];

    newcount = count - tmp;
    tail = (tail + tmp) & (MAX_INBUF_LEN - 1);
    for (i = 0; i < newcount; ++i)
        tmp_buf[tmp + i] = cardscanner_inbuf.buf[tail + i];

    tail = (tail + newcount) & (MAX_INBUF_LEN - 1);

    cardscanner_inbuf.head = head;
    cardscanner_inbuf.tail = tail;
    spin_unlock_irqrestore(&buf_lock, flags);

	if (copy_to_user(buf, tmp_buf, count))
		return -EFAULT;
	return count;
}

static int cardscanner_close(struct inode *inode, struct file *filp)
{
	return 0;
}

static struct file_operations cardscanner_fops = {
	.open = cardscanner_open,
	.read = cardscanner_read,
	.release = cardscanner_close
};

static struct usb_class_driver cardscanner_class_driver = {
	.name = "cardscanner",
	.fops = &cardscanner_fops
};


struct cardscannerusb {
	struct usb_device *usbdev;
	unsigned char old[8];
	struct urb *irq;

	unsigned char *new;
};

static char scan_codes_to_ascii(unsigned char a[8])
{
	return a[0]==2 ? sm[a[2]] : m[a[2]];
}

static void cardscannerusb_irq(struct urb *urb)
{
	struct cardscannerusb* tst = urb->context;
    unsigned long flags;
    int head, tail;

	switch (urb->status) {
		case 0:			/* success */
			break;
		case -ECONNRESET:	/* unlink */
		case -ENOENT:
		case -ESHUTDOWN:
			return;
			/* -EPIPE:  should clear the halt */
		default:		/* error */
			goto resubmit;
	}

    spin_lock_irqsave(&buf_lock, flags);
    head = cardscanner_inbuf.head;
    tail = cardscanner_inbuf.tail;

	if (tst->new[2] != 0 && CIRC_SPACE(head, tail, MAX_INBUF_LEN) > 0) {
        cardscanner_inbuf.buf[head] = scan_codes_to_ascii(tst->new);
        head = (head + 1) & (MAX_INBUF_LEN - 1);
	}

    cardscanner_inbuf.head = head;
    cardscanner_inbuf.tail = tail;
    spin_unlock_irqrestore(&buf_lock, flags);

resubmit:
	if (usb_submit_urb (urb, GFP_ATOMIC))
		printk("Unable to resubmit urb\n");
}

static int cardscannerusb_alloc_mem(struct usb_device *dev, struct cardscannerusb *tstusb)
{
	if (!(tstusb->irq = usb_alloc_urb(0, GFP_KERNEL)))
		return -1;
	if (!(tstusb->new = kmalloc(8, GFP_KERNEL)))
		return -1;

	return 0;
}

static void cardscannerusb_free_mem(struct usb_device *dev, struct cardscannerusb *tst)
{
	usb_free_urb(tst->irq);
	kfree(tst->new);
}

static void cardscanner_disconnect(struct usb_interface *intf)
{
	struct cardscannerusb *tst = usb_get_intfdata (intf);

	usb_deregister_dev(intf, &cardscanner_class_driver);
	usb_set_intfdata(intf, NULL);
	if (tst) {
		usb_kill_urb(tst->irq);
		cardscannerusb_free_mem(interface_to_usbdev(intf), tst);
		kfree(tst);
	}
}

static int cardscanner_probe(struct usb_interface *iface, const struct usb_device_id *id)
{
	struct usb_device *dev = interface_to_usbdev(iface);
	struct usb_host_interface *interface;
	struct usb_endpoint_descriptor *endpoint;
	struct cardscannerusb *tstusb;
	int pipe, maxp;
	int error = -ENOMEM;

	interface = iface->cur_altsetting;

	if (interface->desc.bInterfaceClass != USB_INTERFACE_CLASS_HID ||
	    interface->desc.bInterfaceSubClass != USB_INTERFACE_SUBCLASS_BOOT ||
	    interface->desc.bInterfaceProtocol != USB_INTERFACE_PROTOCOL_KEYBOARD) {
		return -ENODEV;
	}

	if (interface->desc.bNumEndpoints != 1) {
		return -ENODEV;
	}

	endpoint = &interface->endpoint[0].desc;
	if (!usb_endpoint_is_int_in(endpoint)) {
		return -ENODEV;
	}

	pipe = usb_rcvintpipe(dev, endpoint->bEndpointAddress);
	maxp = usb_maxpacket(dev, pipe, usb_pipeout(pipe));

	tstusb = kzalloc(sizeof(struct cardscannerusb), GFP_KERNEL);
	if (!tstusb)
		goto fail1;

	if (cardscannerusb_alloc_mem(dev, tstusb))
		goto fail2;

	tstusb->usbdev = dev;

	usb_register_dev(iface, &cardscanner_class_driver);

	usb_fill_int_urb(tstusb->irq, dev, pipe,
			tstusb->new, (maxp > 8 ? 8 : maxp),
			cardscannerusb_irq, tstusb, endpoint->bInterval);

	usb_set_intfdata(iface, tstusb);
	if (usb_submit_urb(tstusb->irq, GFP_KERNEL))
		return -EIO;

	return 0;

fail2:	
	cardscannerusb_free_mem(dev, tstusb);
fail1:	
	kfree(tstusb);
	return error;
}

static int __init cardscanner_init(void)
{
    int result;

    cardscanner_inbuf.buf = kmalloc(sizeof(char) * MAX_INBUF_LEN, GFP_KERNEL);
    if (!cardscanner_inbuf.buf)
        return -ENOMEM;

	result = usb_register(&cardscanner_driver);
    if (result < 0)
        kfree(cardscanner_inbuf.buf);

	return result;
}

static void __exit cardscanner_exit(void)
{
    kfree(cardscanner_inbuf.buf);
	usb_deregister(&cardscanner_driver);
}

module_init(cardscanner_init);
module_exit(cardscanner_exit);
